/* $Id: pthread_shared_destroy.c,v 1.5 2016/02/17 14:35:19 kostik Exp kostik $ */

#include <sys/mman.h>
#include <errno.h>
#include <pthread.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

static char *sh_page;
static pthread_barrier_t *b;

#define MTX_PER_PROC	10
#define	NPROC		3

static void
child(int idx)
{
	pthread_mutex_t *m;
	pthread_mutexattr_t m_attr;
	int error, i;

	error = pthread_mutexattr_init(&m_attr);
	if (error != 0) {
		fprintf(stderr, "pthread_mutexattr_init failed %s\n",
		    strerror(error));
		exit(1);
	}
	error = pthread_mutexattr_setpshared(&m_attr, 1);
	if (error != 0) {
		fprintf(stderr, "pthread_mutexattr_setpshared failed %s\n",
		    strerror(error));
		exit(1);
	}
	for (i = 0; i < MTX_PER_PROC; i++) {
		m = (pthread_mutex_t *)(sh_page + sizeof(pthread_barrier_t) +
		    sizeof(pthread_mutex_t) * (idx * MTX_PER_PROC + i));
		error = pthread_mutex_init(m, &m_attr);
		if (error != 0) {
			fprintf(stderr, "pthread_mutex_init failed %s\n",
			    strerror(error));
			exit(1);
		}
	}
	pthread_mutexattr_destroy(&m_attr);
	error = pthread_barrier_wait(b);
	if (error != 0 && error != PTHREAD_BARRIER_SERIAL_THREAD) {
		fprintf(stderr, "pthread_barrier_wait failed %s\n",
		    strerror(error));
		exit(1);
	}
	for (i = 0; i < MTX_PER_PROC * NPROC; i++) {
		m = (pthread_mutex_t *)(sh_page + sizeof(pthread_barrier_t) +
		    sizeof(pthread_mutex_t) * i);
		error = pthread_mutex_lock(m);
		if (error != 0) {
			fprintf(stderr, "pthread_mutex_lock failed %s\n",
			    strerror(error));
			exit(1);
		}
		error = pthread_mutex_unlock(m);
		if (error != 0) {
			fprintf(stderr, "pthread_mutex_unlock failed %s\n",
			    strerror(error));
			exit(1);
		}
	}
	error = pthread_barrier_wait(b);
	if (error != 0 && error != PTHREAD_BARRIER_SERIAL_THREAD) {
		fprintf(stderr, "pthread_barrier_wait 2 failed %s\n",
		    strerror(error));
		exit(1);
	}
	for (i = 0; i < MTX_PER_PROC; i++) {
		m = (pthread_mutex_t *)(sh_page + sizeof(pthread_barrier_t) +
		    sizeof(pthread_mutex_t) * ((NPROC - idx) * MTX_PER_PROC +
		    i));
		error = pthread_mutex_destroy(m);
		if (error != 0) {
			fprintf(stderr, "pthread_mutex_destroy failed %s\n",
			    strerror(error));
			exit(1);
		}
	}
	_exit(0);
}

int
main(void)
{
	pthread_barrierattr_t b_attr;
	pid_t p[NPROC];
	int error, i, pid;

	sh_page = mmap(NULL, sizeof(pthread_barrier_t) +
	    sizeof(pthread_mutex_t) * MTX_PER_PROC * NPROC,
	    PROT_READ | PROT_WRITE, MAP_SHARED | MAP_ANON, -1, 0);
	if (sh_page == MAP_FAILED) {
		fprintf(stderr, "mmap failed %s\n", strerror(errno));
		return (1);
	}

	error = pthread_barrierattr_init(&b_attr);
	if (error != 0) {
		fprintf(stderr, "pthread_barrierattr_init failed %s\n",
		    strerror(error));
		return (1);
	}
	error = pthread_barrierattr_setpshared(&b_attr, 1);
	if (error != 0) {
		fprintf(stderr, "pthread_barrierattr_setpshared failed %s\n",
		    strerror(error));
		return (1);
	}
	b = (pthread_barrier_t *)sh_page;
	error = pthread_barrier_init(b, &b_attr, NPROC);
	if (error != 0) {
		fprintf(stderr, "pthread_barrier_init failed %s\n",
		    strerror(error));
		return (1);
	}
	pthread_barrierattr_destroy(&b_attr);

	fflush(stderr);
	for (i = 0; i < NPROC; i++) {
		pid = fork();
		if (pid == -1) {
			fprintf(stderr, "fork failed %s\n", strerror(errno));
			return (1);
		} else if (pid == 0)
			child(i);
		else
			p[i] = pid;
	}
	return (0);
}

